import json
import numpy as np
import pandas as pd
import rasterio
import rasterio.features
from scipy import signal
from shapely import affinity
from shapely.geometry import Polygon
from pathlib import Path
import geopandas as gpd


def convert_shape_to_wgs84(poly, transform):
    a,b,c,d,e,f =  transform[:6]
    mat = (a,b,d,e,c,f)
    return affinity.affine_transform(poly, mat)

def get_vessel_shape(center, width, length, theta):
    """ Get the bounding box for a ship
    Args:
        center (float): Ship's center index (subpixel)
        width (float): Ship's width in pixel
        length (float): Ship's length in pixel
        theta (float): Ship's heading angle in radian
    Returns:
        (Shapely Polygon): Bounding box 
    """
    width = width/2
    length = length/2
    poly = Polygon([(-width, -length), (-width, length), (width, length),
                    (width, -length)])

    poly = affinity.rotate(poly, theta, 'center',  use_radians=True)
    poly = affinity.translate(poly, xoff=center[0], yoff=center[1])
    return poly


def center_cross_correlation(data, width, length, theta, upsample_factor):
    """ Approximate the center of the boat with a cross correlation of the data
    and a gaussian model
    Args:
        data (Rasterio DatasetReader): tif file reader
        width (float): Ship's width in pixel
        length (float): Ship's length in pixel
        theta (float): Ship's heading angle in radian
        upsample_factor (uint): up sampling factor
    Returns:
        (Numpy Array):  centered mask
        (Numpy Array):  Correlations with each bands of the image
        (Numpy Array):  Coordinates of the center of the boat
    """
    # Up sampling
    up_sampled_data = np.repeat(np.repeat(data, upsample_factor, axis=1),
                                    upsample_factor, axis=0)
  
    # Rectangular mask
    size = (data.shape[0]*upsample_factor, data.shape[1]*upsample_factor)
    shape = get_vessel_shape([size[0]/2, size[1]/2], width * upsample_factor, length*upsample_factor, theta)
    centered_mask = rasterio.features.rasterize([shape], out_shape=(size), all_touched=True).astype(np.int8)

    # Remove null edges of the mask
    border_padding_size = 2 * upsample_factor
    max_x = int(np.max(shape.exterior.coords.xy[1])) +3 + border_padding_size
    min_x = int(np.min(shape.exterior.coords.xy[1])) -2 - border_padding_size
    max_y = int(np.max(shape.exterior.coords.xy[0])) +3 + border_padding_size
    min_y = int(np.min(shape.exterior.coords.xy[0])) -2 - border_padding_size

    if (max_x - min_x)%2 == 0:
        max_x += 1
    if (max_y - min_y)%2 == 0:
        max_y += 1

    cropped_mask = centered_mask[min_x:max_x, min_y:max_y]
    background_value = -1
    
    ####
    background_value = 0
    cropped_mask[cropped_mask == 0 ] = -1
    ####
    


    model_mask = cropped_mask.copy()
    model_mask[0:border_padding_size] = background_value
    model_mask[-border_padding_size:] = background_value
    model_mask[:,0:border_padding_size] = background_value
    model_mask[:,-border_padding_size:] = background_value
    for i in range(2,model_mask.shape[0]-2):
        for j in range(2, model_mask.shape[1]-2):
            window = cropped_mask[i-2:i+3,j-2:j+3]
            if np.all(window != 1):
                model_mask[i,j] = background_value

    # Model and data must have the same dimensions parity to avoid an offset after the cross correlation
    if model_mask.shape[0] %2 == 0:
        model_mask = np.append(model_mask, -1* np.ones((1, model_mask.shape[1])), axis=0)
    if model_mask.shape[1] %2 == 0:
        model_mask = np.append(model_mask, -1* np.ones((model_mask.shape[0],1)), axis=1)

    # Compute the coordinates of the maximum correlation for each band
    mean_filter = np.ones(cropped_mask.shape)/np.size(cropped_mask)
    correlations = np.zeros(up_sampled_data.shape)
    for i in range(data.shape[2]):
        correlation_band_i = signal.correlate2d(up_sampled_data[:, :, i],
                                                   model_mask, mode='valid', boundary='fill', fillvalue=0)
        
        # Mean filter convolution (removes abnormal high values)
        average_map = signal.convolve2d(up_sampled_data[:, :, i],
                                        mean_filter, mode='valid', boundary='fill', fillvalue=np.inf)  
        correlation_band_i = np.divide(correlation_band_i, average_map, 
                out=np.zeros(correlation_band_i.shape), where=average_map!=0) 
        correlation_band_i -= np.min(correlation_band_i)

        # resize the output
        nb = correlations.shape[0]
        na = correlation_band_i.shape[0]
        lowerx = (nb) // 2 - (na // 2)
        upperx = (nb // 2) + (na // 2)
        nb = correlations.shape[1]
        na = correlation_band_i.shape[1]
        lowery = (nb) // 2 - (na // 2)
        uppery = (nb // 2) + (na // 2)
        correlations[lowerx:upperx, lowery:uppery,i] = correlation_band_i


    window1d = np.abs(np.hamming(correlations.shape[0]))
    window2d = np.sqrt(np.outer(window1d,window1d))
    correlations *= np.repeat(window2d[:,:,None],correlations.shape[2],axis=2)

    somme = np.sum(correlations, axis = 2)
    
    ##########################################################

    

    center =  np.argwhere(somme == np.max(somme))[0] / upsample_factor

    # swap to the right coordinate system
    center[0], center[1] = center[1], center[0]

    return centered_mask, correlations, center


def generate_mask(file_path, output_dir, row_vessel, upsample_factor=1, output_type="raster", dilatation=5):
    """ Get the entries for the csv file for an image without a mask
    Args:
        file_path (String): path to the image
        output_dir (String): path to the output folder
        row_vessel (Pandas Dataframe): boat's informations
        upsample_factor (Int): upsampling factor for the model and the cross correlation
        output_type (String): can be a 'raster' (.tif) or a 'vector' (.geojson) output
        dilatation (Int): size of the dilation for the length and width of the mask in meters
    Returns:
        (Dict): csv entries for the image  
    """
    
    #  Load the source tile
    with rasterio.open(file_path, 'r') as src:

        # All 10m channels except blue
        data = src.read([2,3,4,11,12]).transpose(1, 2, 0)
        
        # Get the dimensions of the boat
        # pixel_res = abs(src.transform[4])
        pixel_res = 10
        width_px = row_vessel["Width"] / pixel_res
        length_px = row_vessel["Length"] / pixel_res

        # Get the center of the mask
        model, correlations, center = center_cross_correlation(data,
                width_px, length_px, row_vessel['Heading'], upsample_factor)

        # Rectangular bounding box
        mask_width_px = (row_vessel["Width"] + dilatation) / pixel_res
        mask_length_px = (row_vessel["Length"] + dilatation) / pixel_res
        shape = get_vessel_shape(center, mask_width_px, mask_length_px, row_vessel['Heading'])
        mask = rasterio.features.rasterize([shape], out_shape=data.shape[0:2], all_touched=True)

        if output_type == "vector":
            # generate Geojson  
            shape_wgs84 = convert_shape_to_wgs84(shape, src.transform)
            gdr = gpd.GeoDataFrame({'geometry': [shape_wgs84],  
                                "labels": [["Boat"]], 
                                "tasks":[["segmentation"]]}, 
                                crs='EPSG:4326')

            # Write to geojson
            label_geojson = gdr.to_json(drop_id=True)
            out_mask_path = Path(file_path).parent / f"{file_path.stem}_labels.geojson"

            with open(out_mask_path, "w") as dst:
                dst.write(label_geojson)

        elif output_type == "raster":
            # Create the profile for the .tif
            profile = {
                'driver': 'GTiff',
                'height': src.height,
                'width': src.width,
                'count': 1,
                'dtype': rasterio.uint8,
                'crs': src.crs,
                'transform': src.transform,
                'nodata': None,
                'compress': 'lzw'
            }

            # Write the mask
            out_mask_path = Path(output_dir) / f"mask_{Path(file_path).name}"
            with rasterio.open(out_mask_path, 'w', **profile) as dst:
                dst.write(mask, indexes=1)


def process_directoy(dir_name, output_dir, tiles_df, upsample_factor, output_type):

    assert output_type == "raster" or output_type == "vector"
    print(f"Processing tiles from {dir_name}")

    # Prepare outputpath
    output_dir_name = Path(output_dir)

    if output_type =="vector":

        labels_json = {"labels": [{"name": "Boat", "color": "#ff8000"}]}
        json_path = Path(dir_name) / "labels.json"

        with open(json_path, "w") as dst:
            dst.write(json.dumps(labels_json))

    # Generate the mask for each vessel
    list_nan = []

    # Convert angles in radian
    tiles_df["Heading"] *= np.pi/180

    for i in range(len(tiles_df)):
        row_vessel = tiles_df.iloc[i]
        img_path = Path(dir_name) / f"{row_vessel['ImageId']}.tiff"

        # Get mmsi from filename
        vessel = int(row_vessel["ImageId"].split('_')[1])
        
        if row_vessel.isnull().values.any(): # Tiles with Nan data
            list_nan.append(vessel)
        
        else:
            # Get the mask of the boat
            generate_mask(img_path, output_dir_name, row_vessel, upsample_factor, output_type)
        
    if list_nan:
        print(f"No mask saved for the vessels {list_nan}, NaN values found in the AIS data")


def process_db(images_path, output_dir, upsample_factor, output_type = "raster"):
    """ Get the entries for the csv file for an image without a mask
    Args:
        csv_tiles (String): path to the csv folder of the tiles
        output_dir (String): path to the output folder
        upsample_factor (Int): upsampling factor for the model and the cross correlation
        output_type (String): can be a 'raster' (.tif) or a 'vector' (.geojson) output
    """
    # Prepare the output file
    csv_paths = Path(images_path).glob("*.csv")
    for csv_tiles in csv_paths:
        # read the tiles csv
        tiles_df = pd.read_csv(csv_tiles)
        # Get the path to the tiles
        tiles_path = Path(csv_tiles).parent
        process_directoy(tiles_path,
                         output_dir=output_dir, 
                         tiles_df=tiles_df, 
                         upsample_factor=upsample_factor,
                         output_type=output_type)



import typer
from pathlib import Path

from ..fe import (
    ingest_openeo,
    retrieve_pipelines,
    stage_pipeline,
    deactivate_pipeline,
)

app = typer.Typer(help="Explore, ingest and download Feature Engineering Pipelines.")

@app.command()
def ingest(
    path: Path = typer.Option(..., "--path", "-p", help="Path to the pipeline to ingest"),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="Verbose output. This will print the progress of the ingestion",
    ),
    foce_metadata_update: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force metadata update even if it already exists. Will overwrite the current metadata in EOTDL",
    ),
    sync_metadata: bool = typer.Option(
        False,
        "--sync",
        "-s",
        help="Sync local metadata with the EOTDL. Will overwrite the local metadata",
    ),
):
    """
    Ingest a pipeline to the EOTDL.

    This command ingests the pipeline to the EOTDL. The pipeline must be a folder with the pipeline files,
    and at least a README.md file. All the files in the folder will be uploaded to the EOTDL.
    \n\n
    Examples\n
    --------\n
    $ eotdl pipelines ingest --path /path/to/folder-with-pipeline --verbose True
    """
    try:
        ingest_openeo(path, verbose, typer.echo, foce_metadata_update, sync_metadata)
    except Exception as e:
        typer.echo(e)

@app.command()
def list(
    name: str = typer.Option(
        None, "--name", "-n", help="Filter the returned pipelines by name"
    ),
    limit: int = typer.Option(
        None, "--limit", "-l", help="Limit the number of returned results"
    ),
):
    """
    Retrieve a list with all the pipelines in the EOTDL.

    If using --name, it will filter the results by name. If no name is provided, it will return all the pipelines.\n
    If using --limit, it will limit the number of results. If no limit is provided, it will return all the pipelines.
    \n\n
    Examples\n
    --------\n
    $ eotdl pipelines list\n
    $ eotdl pipelines list --name YourPipeline --limit 5
    """
    try:
        pipelines = retrieve_pipelines(name, limit)
        typer.echo(pipelines)
    except Exception as e:
        typer.echo(e)

@app.command()
def get(
    pipeline: str = typer.Argument(None, help="Name of the pipeline to download"),
    path: str = typer.Option(
        None, "--path", "-p", help="Download the pipeline to a specific output path"
    ),
    version: int = typer.Option(None, "--version", "-v", help="pipeline version"),
    assets: bool = typer.Option(
        False, "--assets", "-a", help="Download STAC assets from the pipeline"
    ),
    force: bool = typer.Option(
        False, "--force", "-f", help="Force download even if file exists"
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="Verbose output. This will print the progress of the download",
    ),
):
    """
    Download a pipeline from the EOTDL.
    \n\n
    If using --path, it will download the pipeline to the specified path. If no path is provided, it will download to ~/.eotdl/pipelines.\n
    If using --version, it will download the specified version. If no version is provided, it will download the latest version.\n
    If using --assets when the pipeline is STAC, it will also download the STAC assets of the pipeline. If not provided, it will only download the STAC metadata.\n
    If using --force, it will download the pipeline even if the file already exists.\n
    If using --verbose, it will print the progress of the download.
    \n\n
    Examples\n
    --------\n
    $ eotdl pipelines get Yourpipeline\n
    $ eotdl pipelines get Yourpipeline --path /path/to/download  --version 1 --assets True --force True --verbose True
    """
    try:
        dst_path = stage_pipeline(
            pipeline, version, path, typer.echo, assets, force, verbose
        )
        typer.echo(f"Data available at {dst_path}")
    except Exception as e:
        typer.echo(e)


@app.command()
def delete(
    pipeline_name: str = typer.Argument(None, help="Name of the pipeline to delete")
):
    """
    Delete a model from the EOTDL.
    """
    try:
        deactivate_pipeline(pipeline_name)
        typer.echo(f"Pipeline {pipeline_name} deleted")
    except Exception as e:
        typer.echo(e)


if __name__ == "__main__":
    app()



import os
import requests

import pytest
import pystac

from jsonschema import validate, RefResolver

from .fixtures import tmp_sentinel_2, sentinel_2

from eotdl.curation.stac.stac import STACGenerator
from eotdl.curation.stac.assets import STACAssetGenerator, BandsAssetGenerator
from eotdl.curation.stac.parsers import UnestructuredParser
from eotdl.curation.stac.dataframe_labeling import LabeledStrategy, UnlabeledStrategy

CATALOG_SCHEMA_URL = 'https://github.com/radiantearth/stac-spec/blob/master/catalog-spec/json-schema/catalog.json'
COLLECTION_SCHEMA_URL = 'https://github.com/radiantearth/stac-spec/blob/master/collection-spec/json-schema/collection.json'


@pytest.mark.parametrize("labeling_strategy", [
    (LabeledStrategy),
    (UnlabeledStrategy)
])
def test_stac_generation_labeling(labeling_strategy, sentinel_2):
    stac_generator = STACGenerator(item_parser=UnestructuredParser, 
                                   assets_generator=STACAssetGenerator, 
                                   labeling_strategy=labeling_strategy,
                                   image_format='tif'
                                   )
    df = stac_generator.get_stac_dataframe(sentinel_2)
    stac_generator.generate_stac_metadata(stac_id='test-dataset',
                                          description='Test dataset',
                                          output_folder='tmp/sentinel_2_stac')
    
    assert_paths_exists('tmp/sentinel_2_stac')
    
    catalog = pystac.Catalog.from_file('tmp/sentinel_2_stac/catalog.json')
    collection = pystac.Collection.from_file('tmp/sentinel_2_stac/source/collection.json')
    assert_catalog_collection_schema(catalog, collection)


def test_stac_generation_bands(tmp_sentinel_2):
    stac_generator = STACGenerator(item_parser=UnestructuredParser, 
                                   assets_generator=BandsAssetGenerator, 
                                   labeling_strategy=LabeledStrategy,
                                   image_format='tif'
                                   )
    bands = {'Boadella': ['B02', 'B03', 'B04'],}
    df = stac_generator.get_stac_dataframe(tmp_sentinel_2, bands=bands)
    stac_generator.generate_stac_metadata(stac_id='test-dataset',
                                          description='Test dataset',
                                          output_folder='tmp/sentinel_2_stac')
    
    assert_paths_exists('tmp/sentinel_2_stac')

    catalog = pystac.Catalog.from_file('tmp/sentinel_2_stac/catalog.json')
    collection = pystac.Collection.from_file('tmp/sentinel_2_stac/source/collection.json')
    assert_catalog_collection_schema(catalog, collection)

    assert len(os.listdir(tmp_sentinel_2)) == 10

    items = collection.get_all_items()
    for i in items:
        assets = i.to_dict()["assets"]
        assert len(assets) == 3


def assert_paths_exists(dir: str):
    """
    Assert that all the paths in the directory exist
    """
    assert os.path.exists(f'{dir}/catalog.json')
    assert os.path.exists(f'{dir}/source/collection.json')
    assert len(os.listdir(f'{dir}/source')) == 3


def assert_catalog_collection_schema(catalog: pystac.Catalog, collection: pystac.Collection):
    """
    Assert that the catalog and collection are valid against the schema
    """
    schema = requests.get(CATALOG_SCHEMA_URL).json()
    resolver = RefResolver(base_uri=CATALOG_SCHEMA_URL, referrer=schema)
    
    assert validate(instance=catalog.to_dict(), schema=schema, resolver=resolver) is None

    schema = requests.get(COLLECTION_SCHEMA_URL).json()
    resolver = RefResolver(base_uri=COLLECTION_SCHEMA_URL, referrer=schema)

    assert validate(instance=collection.to_dict(), schema=schema, resolver=resolver) is None



"""
Module for data engineeringt
"""

import re
import tarfile
import datetime
import json
from os.path import exists
from typing import Union, Optional
import geopandas as gpd
import pandas as pd

from shapely.geometry import box
from .geo_utils import get_image_bbox


def get_images_by_location(gdf: gpd.GeoDataFrame, column: str) -> pd.DataFrame:
    """
    Generate a GeoDataFrame with the available images for each location in the dataset.

    :param gdf: GeoDataFrame generated from the ItemCollection of a sen12floods collection
    :return gdf_dates_per_aoi:  GeoDataFrame with the available images for each location in
                                the dataset.
            The returned GeoDataFrame has three columns:
                - location_id: the unique ID of each location.
                - images_count: the count of available images of each location.
                - images_dates: list with the dates of the available images of each location.
    """
    uniques_location_id = gdf[column].unique()  # List of unique location ids
    uniques_location_id.sort()

    images_count_list, images_dates_list = [], []

    # Iterate the unique location ids, count the number of images per location and generate
    # a list with the dates of every image in a location
    for location_id in uniques_location_id:
        dates = gdf[gdf[column] == location_id]["datetime"]
        images_count_list.append(dates.count())
        images_dates_list.append(dates.tolist())

    images_dates_list.sort()  # Sort the list of dates
    data = {
        column: uniques_location_id,
        "dates_count": images_count_list,
        "dates_list": images_dates_list,
    }
    df_dates_per_aoi = pd.DataFrame.from_dict(data)

    return df_dates_per_aoi


def generate_location_payload(
    gdf: Union[gpd.GeoDataFrame, pd.DataFrame], path: str
) -> dict:
    """
    Generate a dictionary with the location payload of the locations in the GeoDataFrame,
    such as the bounding box and the time interval to search for available data.
    """
    payload_cache = f"{path}/location_payload.json"
    if exists(payload_cache):
        # Read as dict
        with open(payload_cache, "r", encoding="utf-8") as f:
            payload = json.load(f)
        return payload

    bbox_date_by_location = {}
    for _, row in gdf.iterrows():
        # Get list from dates_list column
        dates_list = list(row["dates_list"])
        for date in dates_list:
            location_id = row["location_id"]
            date_formatted = datetime.datetime.strptime(
                date, "%Y-%m-%dT%H:%M:%S.%fZ"
            ).strftime("%Y-%m-%d")
            location_id_formatted = f"{location_id}_{date_formatted}"
            bbox_date_by_location[location_id_formatted] = {
                "bounding_box": row["geometry"].bounds,
                # Convert str to datetime
                "time_interval": (date, date),
            }

    # Save to json
    with open(payload_cache, "w", encoding="utf-8") as f:
        json.dump(bbox_date_by_location, f)

    return bbox_date_by_location


def get_tarfile_image_info(
    tar: str,
    path: Optional[str] = None,
    pattern: Optional[str] = r"\d{8}T\d{6}",
    level: Optional[int] = 2,
):
    """
    Generate a GeoDataFrame with the available images for each location in the dataset.
    """
    if path:
        gdf_cache = f"{path}/tarfile_images_info.csv"
        if exists(gdf_cache):
            images_gdf = gpd.read_file(
                gdf_cache, GEOM_POSSIBLE_NAMES="geometry", KEEP_GEOM_COLUMNS="NO"
            )
            images_gdf.set_crs(epsg=4326, inplace=True)

            return images_gdf

    images_df = pd.DataFrame()
    with tarfile.open(tar, "r:gz") as tarf:
        rasters = [
            i for i in tarf.getnames() if i.endswith(".tif") or i.endswith(".tiff")
        ]
        for raster in rasters:
            r = tarf.extractfile(raster)
            bbox = get_image_bbox(r)
            date = extract_image_date_in_folder(raster, pattern)
            image_id = extract_image_id_in_folder(raster, level)
            # Use pd.concat to append to dataframe
            images_df = pd.concat(
                [
                    images_df,
                    pd.DataFrame(
                        {"location_id": [image_id], "datetime": [date], "bbox": [bbox]}
                    ),
                ]
            )

    # Clean duplicates
    images_df = images_df.drop_duplicates(subset=["location_id", "datetime"])
    # Convert to geodataframe
    images_gdf = gpd.GeoDataFrame(
        images_df,
        crs="EPSG:4326",
        geometry=images_df["bbox"].apply(lambda x: box(x[0], x[1], x[2], x[3])),
    )
    # Drop bbox column
    images_gdf = images_gdf.drop(columns=["bbox"])
    # Set crs
    images_gdf = images_gdf.set_crs(epsg=4326)
    # Sort by location_id
    images_gdf = images_gdf.sort_values(by=["location_id"])
    if path:
        # Save to csv
        images_gdf.to_csv(gdf_cache, index=False)

    return images_gdf


def extract_image_date_in_folder(raster_path: str, pattern: str):
    """
    Extract the date from the folder name of the image.
    """
    case = re.findall(pattern, raster_path)

    if case:
        date = case[0]
        # Convert date to format YYYY-MM-DDT00:00:00.000Z as datetime object
        formatted_date = datetime.datetime.strptime(date, "%Y%m%dT%H%M%S").strftime(
            "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        return formatted_date

    return None


def extract_image_id_in_folder(raster_path: str, level: int):
    """
    Extract the location id from the folder name of the image, given the level of the folder.
    """
    return raster_path.split("/")[level]


def format_product_location_payload(
    location_payload: dict, images_response: dict, all_info: bool = False
) -> dict:
    """
    Format the location payload with the images response.
    """
    for _, _ in location_payload.items():
        # Add new key to the dictionary
        if all_info:
            location_payload[id]["image"] = (
                images_response[id] if id in images_response else None
            )
        else:
            location_payload[id]["image"] = (
                images_response[id]["properties"]["id"] if images_response[id] else None
            )

    return location_payload



"""
Utils
"""

import requests


def get_airbus_access_token(api_key: str) -> str:
    """
    Get Airbus access token

    Returns:
      str: access token
    """
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
    }

    data = [
        ("apikey", api_key),
        ("grant_type", "api_key"),
        ("client_id", "IDP"),
    ]

    response = requests.post(
        "https://authenticate.foundation.api.oneatlas.airbus.com/auth/realms/IDP/protocol/openid-connect/token",
        headers=headers,
        data=data,
        timeout=60
    )

    access_token = response.json()["access_token"]

    return access_token



"""
Module for data engineering with STAC elements
"""

from os.path import dirname, join, abspath
from os import makedirs
from json import dumps
from typing import Union, Optional
from shutil import rmtree
from traceback import print_exc

import geopandas as gpd
import pystac

from tqdm import tqdm


def stac_items_to_gdf(items: pystac.ItemCollection) -> gpd.GeoDataFrame:
    """
    Get a GeoDataFrame from a given pystac.ItemCollection.

    :param: items: A pystac.ItemCollection
    :return: GeoDataframe from the given ItemCollection
    """
    _features = [i.to_dict() for i in items]

    # Get a new ItemCollection by removing duplicate items, if they exist
    features = []
    for f in _features:
        if f not in features:
            # Add all the keys in the properties dict as columns in the GeoDataFrame
            for k, v in f.items():
                if k not in f["properties"] and k != "geometry":
                    f["properties"][k] = v
            if "scene_id" in f["properties"]:
                f["properties"]["scene_id"] = f["id"].split("_")[3]
            features.append(f)

    return gpd.GeoDataFrame.from_features(features)


def get_all_children(obj: pystac.STACObject) -> list:
    """
    Get all the children of a STAC object
    """
    children = []
    # Append the current object to the list
    children.append(obj.to_dict())

    # Collections
    collections = list(obj.get_collections())
    for collection in collections:
        children.append(collection.to_dict())

    # Items
    items = obj.get_items()
    for item in items:
        children.append(item.to_dict())

    # Items from collections
    for collection in collections:
        items = collection.get_items()
        for item in items:
            children.append(item.to_dict())

    return children


def make_links_relative_to_path(
    path: str,
    catalog: Union[pystac.Catalog, str],
) -> pystac.Catalog:
    """
    Makes all asset HREFs and links in the STAC catalog relative to a given path
    """
    if isinstance(catalog, str):
        catalog = pystac.read_file(catalog)
    path = abspath(path)

    # Create a temporary catalog in the destination path to set as root
    future_path = join(path, "catalog.json")
    makedirs(path, exist_ok=True)
    with open(future_path, "w", encoding="utf-8") as f:
        f.write(dumps(catalog.to_dict(), indent=4))
    temp_catalog = pystac.Catalog.from_file(future_path)

    catalog.set_root(temp_catalog)
    catalog.make_all_asset_hrefs_absolute()

    for collection in catalog.get_children():
        # Create new collection
        new_collection = collection.clone()
        new_collection.set_self_href(join(path, collection.id, "collection.json"))
        new_collection.set_root(catalog)
        new_collection.set_parent(catalog)
        # Remove old collection and add new one to catalog
        catalog.remove_child(collection.id)
        catalog.add_child(new_collection)
        for item in collection.get_all_items():
            # Create new item from old collection and add it to the new collection
            new_item = item.clone()
            new_item.set_self_href(
                join(path, collection.id, item.id, f"{item.id}.json")
            )
            new_item.set_parent(collection)
            new_item.set_root(catalog)
            new_item.make_asset_hrefs_relative()
            new_collection.add_item(new_item)

    catalog.make_all_asset_hrefs_relative()

    return catalog


def merge_stac_catalogs(
    catalog_1: Union[pystac.Catalog, str],
    catalog_2: Union[pystac.Catalog, str],
    destination: Optional[str] = None,
    keep_extensions: Optional[bool] = False,
    catalog_type: Optional[pystac.CatalogType] = pystac.CatalogType.SELF_CONTAINED,
) -> None:
    """
    Merge two STAC catalogs, keeping the properties, collection and items of both catalogs
    """
    if isinstance(catalog_1, str):
        catalog_1 = pystac.Catalog.from_file(catalog_1)
    if isinstance(catalog_2, str):
        catalog_2 = pystac.Catalog.from_file(catalog_2)

    for col1 in tqdm(catalog_1.get_children(), desc="Merging catalogs..."):
        # Check if the collection exists in catalog_2
        col2 = catalog_2.get_child(col1.id)
        if not col2:
            # If it does not exist, add it
            col1_ = col1.clone()
            catalog_2.add_child(col1)
            col2 = catalog_2.get_child(col1.id)
            col2.clear_items()
            for i in col1_.get_stac_objects(pystac.RelType.ITEM):
                col2.add_item(i)
        else:
            # If it exists, merge the items
            for item1 in col1.get_items():
                if col2.get_item(item1.id) is None:
                    col2.add_item(item1)

    if keep_extensions:
        for ext in catalog_1.stac_extensions:
            if ext not in catalog_2.stac_extensions:
                catalog_2.stac_extensions.append(ext)

        for extra_field_name, extra_field_value in catalog_1.extra_fields.items():
            if extra_field_name not in catalog_2.extra_fields:
                catalog_2.extra_fields[extra_field_name] = extra_field_value

    if destination:
        make_links_relative_to_path(destination, catalog_2)
    else:
        destination = dirname(catalog_2.get_self_href())

    # Save the merged catalog
    try:
        print("Validating and saving...")
        catalog_2.validate()
        rmtree(
            destination
        ) if not destination else None  # Remove the old catalog and replace it with the new one
        catalog_2.normalize_and_save(root_href=destination, catalog_type=catalog_type)
        print("Success!")
    except pystac.STACValidationError:
        # Return full callback
        print_exc()



import typer
from pathlib import Path

from ..models import (
    retrieve_models,
    ingest_model,
    stage_model,
    deactivate_model
)

app = typer.Typer(help="Explore, ingest and download ML models.")

@app.command()
def ingest(
    path: Path = typer.Option(..., "--path", "-p", help="Path to the model to ingest"),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="Verbose output. This will print the progress of the ingestion",
    ),
    foce_metadata_update: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force metadata update even if it already exists. Will overwrite the current metadata in EOTDL",
    ),
    sync_metadata: bool = typer.Option(
        False,
        "--sync",
        "-s",
        help="Sync local metadata with the EOTDL. Will overwrite the local metadata",
    ),
):
    """
    Ingest a model to the EOTDL.

    This command ingests the model to the EOTDL. The model must be a folder with the model files,
    and at least a metadata.yml file or a catalog.json file. If there are not these files, the ingestion
    will not work. All the files in the folder will be uploaded to the EOTDL.
    \n\n
    The following constraints apply to the model name:\n
    - It must be unique\n
    - It must be between 3 and 45 characters long\n
    - It can only contain alphanumeric characters and dashes.\n
    \n
    The metadata.yml file must contain the following fields:\n
    - name: the name of the model\n
    - authors: the author or authors of the model\n
    - license: the license of the model\n
    - source: the source of the model\n
    - thumbnail: an image to use as the thumbnail of the dataset in the website\n
    \n
    If using --verbose, it will print the progress of the ingestion.
    \n\n
    Examples\n
    --------\n
    $ eotdl models ingest --path /path/to/folder-with-model --verbose True
    """
    try:
        ingest_model(path, verbose, typer.echo, foce_metadata_update, sync_metadata)
    except Exception as e:
        typer.echo(e)

@app.command()
def list(
    name: str = typer.Option(
        None, "--name", "-n", help="Filter the returned models by name"
    ),
    limit: int = typer.Option(
        None, "--limit", "-l", help="Limit the number of returned results"
    ),
):
    """
    Retrieve a list with all the models in the EOTDL.

    If using --name, it will filter the results by name. If no name is provided, it will return all the models.\n
    If using --limit, it will limit the number of results. If no limit is provided, it will return all the models.
    \n\n
    Examples\n
    --------\n
    $ eotdl models list\n
    $ eotdl models list --name YourModel --limit 5
    """
    try:
        models = retrieve_models(name, limit)
        typer.echo(models)
    except Exception as e:
        typer.echo(e)

@app.command()
def get(
    model: str = typer.Argument(None, help="Name of the model to download"),
    path: str = typer.Option(
        None, "--path", "-p", help="Download the model to a specific output path"
    ),
    version: int = typer.Option(None, "--version", "-v", help="Model version"),
    assets: bool = typer.Option(
        False, "--assets", "-a", help="Download STAC assets from the model"
    ),
    force: bool = typer.Option(
        False, "--force", "-f", help="Force download even if file exists"
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="Verbose output. This will print the progress of the download",
    ),
):
    """
    Download a model from the EOTDL.
    \n\n
    If using --path, it will download the model to the specified path. If no path is provided, it will download to ~/.eotdl/models.\n
    If using --version, it will download the specified version. If no version is provided, it will download the latest version.\n
    If using --assets when the model is STAC, it will also download the STAC assets of the model. If not provided, it will only download the STAC metadata.\n
    If using --force, it will download the model even if the file already exists.\n
    If using --verbose, it will print the progress of the download.
    \n\n
    Examples\n
    --------\n
    $ eotdl models get YourModel\n
    $ eotdl models get YourModel --path /path/to/download --version 1 --assets True --force True --verbose True
    """
    try:
        dst_path = stage_model(
            model, version, path, typer.echo, assets, force, verbose
        )
        typer.echo(f"Data available at {dst_path}")
    except Exception as e:
        typer.echo(e)


@app.command()
def delete(
    model_name: str = typer.Argument(None, help="Name of the model to delete")
):
    """
    Delete a model from the EOTDL.
    """
    try:
        deactivate_model(model_name)
        typer.echo(f"Model {model_name} deleted")
    except Exception as e:
        typer.echo(e)


if __name__ == "__main__":
    app()



import rasterio as rio
import numpy as np
from torch.utils.data import Dataset
import torch

class SegmentationDataset(Dataset):
    def __init__(self, images, masks, bands=[1,2,3], norm_value = 4000, trans=None, num_classes=6):
        self.images = images
        self.masks = masks
        self.trans = trans
        self.bands = bands
        self.norm_value = norm_value
        self.num_classes = num_classes
        
    def __len__(self):
        return len(self.images)
    
    def __getitem__(self, idx):
        image = rio.open(self.images[idx]).read(self.bands)
        image = np.clip(image / self.norm_value, 0, 1).transpose(1,2,0)
        mask = rio.open(self.masks[idx]).read(1)
        if self.trans is not None:
            trans = self.trans(image=image, mask=mask)
            image, mask = trans['image'], trans['mask']
        return torch.from_numpy(image.transpose(2,0,1)).float(), \
            torch.nn.functional.one_hot(torch.from_numpy(mask).long(), num_classes=self.num_classes).float().permute(2,0,1)
    
class BinarySegmentationDataset(torch.utils.data.Dataset):
    def __init__(self, images, masks, trans=None, bands=(4,3,2), norm_value=4000):
        self.images = images
        self.masks = masks
        self.trans = trans
        self.bands = bands
        self.norm_value = norm_value

    def __len__(self):
        return len(self.images)
    
    def __getitem__(self, idx):
        with rio.Env(CPL_LOG_LEVEL='ERROR'):
            with rio.open(self.images[idx]) as image_src:
                image = np.clip(image_src.read(self.bands) / self.norm_value, 0, 1).astype(np.float32).transpose(1, 2, 0)
            with rio.open(self.masks[idx]) as mask_src:
                mask = (mask_src.read(1) / 255.0).astype(np.float32)
        
        if self.trans:
            trans = self.trans(image=image, mask=mask)
            image, mask = trans['image'].transpose(2, 0, 1), trans['mask']
        return torch.from_numpy(image).float(), torch.from_numpy(mask).float().unsqueeze(0)
    


from .dataframe_utils import *
import openeo
from openeo.extra.job_management import MultiBackendJobManager, CsvJobDatabase
from .s3proxy_utils import upload_geoparquet_file
import geojson

def start_job(row: pd.Series, connection: openeo.Connection, **kwargs) -> openeo.BatchJob:

        temporal_extent = eval(row["temporal_extent"])
        crs = row['crs']

        # set up load url in order to allow non-latlon feature collections for spatial filtering
        geometry = geojson.loads(row["geometry"])
        features = gpd.GeoDataFrame.from_features(geometry).set_crs(crs)
        url = upload_geoparquet_file(features,connection)

        #run the s1 and s2 udp
        s1 = connection.datacube_from_process(
                "s1_weekly_statistics",
                namespace="https://raw.githubusercontent.com/earthpulse/eotdl/refs/heads/hv_openeoexample/tutorials/notebooks/openeo/s1_weekly_statistics.json",
                temporal_extent=temporal_extent,
                )
        
        s2 = connection.datacube_from_process(
                "s2_weekly_statistics",
                namespace="https://raw.githubusercontent.com/earthpulse/eotdl/refs/heads/hv_openeoexample/tutorials/notebooks/openeo/s2_weekly_statistics.json",
                temporal_extent=temporal_extent,
                )
        
        #merge both cubes and filter across the feature collection
        merged = s2.merge_cubes(s1)
        result = merged.filter_spatial(connection.load_url(url, format="Parquet"))
        
        #dedicated job settings to save the individual features within a collection seperately
        job = result.create_job(
                out_format="NetCDF",
                sample_by_feature = True,
                feature_id_property="id",
                filename_prefix = "eotdl"
        )

        return job

def patch_extraction(
        gdf,
        start_date,
        nb_months,
        pixel_size = 64,
        resolution = 10,
        max_points = 5,
        job_tracker = 'jobs.csv',
        parallel_jobs=2
):
        """
        # Transform GeoDataFrame for MultiBackendJobManager

        This function processes an input GeoDataFrame and prepares it for use with openEO's **MultiBackendJobManager**. The job manager enables launching and tracking multiple openEO jobs simultaneously, which is essential for large-scale data extractions. 

        ### Example Use Case
        The function creates patches (e.g., 64x64 pixels) around polygon centers. These patches are suitable for machine learning applications, such as training convolutional neural networks (CNNs).  
        By combining patches into Sentinel-2 grid collections, the workflow ensures cost efficiency and optimized data extraction.

        ### Workflow

        1. **Process the GeoDataFrame**  
        - Create patches with a fixed size around the center of polygon geometries.  
        - Calculate temporal extents for each geometry.  

        2. **Combine Features Using Sentinel-2 Tiling**  
        - Group buffered geometries into collections based on the Sentinel-2 tiling grid.  
        - Minimize redundant openEO cost.  

        3. **Generate Job Metadata DataFrame**  
        - Convert processed data into a DataFrame, ready for the MultiBackendJobManager.

        ### Parameters

        #### Spatial Parameters:
        - **Buffer Distance:** Buffer size (e.g., 320 meters for a 64x64 patch around polygon centers).  
        - **Resolution:** Spatial alignment resolution in meters.

        #### Temporal Parameters:
        - **Start Date:** Start of the temporal extent (e.g., `"2020-01-01"`).  
        - **Number of Months:** Duration of the temporal extent in months.

        #### Job Splitting Parameters:
        - **Max Points Per Job:** Maximum number of features per job batch.
        """
        job_df = process_and_create_advanced_patch_jobs(
                gdf, start_date, nb_months, pixel_size, resolution, max_points=max_points
        )
        """
        # Start Job with Standardized UDPs and Feature Collection Filtering

        This function initializes an openEO batch job using standardized **User-Defined Processes (UDPs)** for Sentinel-1 and Sentinel-2 data processing. It employs a spatial filter designed for non-lat/lon feature collections to ensure precise patch sizes in UTM coordinates.

        ### Key Features

        1. **Use of Standardized UDPs**  
        - **S1 Weekly Statistics:** Computes weekly statistics from Sentinel-1 data.  
        - **S2 Weekly Statistics:** Computes weekly statistics from Sentinel-2 data.  
        - UDPs are defined in external JSON files.

        2. **Spatial Filtering with `load_url`**  
        - Accepts feature collections in **UTM coordinates** to guarantee patches with exact dimensions (e.g., 64x64 meters).  
        - Features are uploaded as a GeoParquet file to an creodias S3 bucket, enabling spatial filtering directly on the server.

        3. **Cube Merging**  
        - Merges Sentinel-1 and Sentinel-2 datacubes for combined analysis.

        4. **Job Configuration**  
        - Saves each feature in the collection as a separate file.  
        - Outputs results in **NetCDF** format with filenames derived
        """
        # Authenticate and add the backend
        connection = openeo.connect(url="openeo.dataspace.copernicus.eu").authenticate_oidc()
        # initialize the job manager
        manager = MultiBackendJobManager()
        manager.add_backend("cdse", connection=connection, parallel_jobs=parallel_jobs)
        job_db = CsvJobDatabase(path=job_tracker)
        if not job_db.exists():
            df = manager._normalize_df(job_df)
            job_db.persist(df)
        manager.run_jobs(start_job=start_job, job_db=job_db)


import torch.nn as nn
import timm
import torch.nn.functional as F
import torch

def conv3x3_bn(ci, co):
    return torch.nn.Sequential(
        torch.nn.Conv2d(ci, co, 3, padding=1),
        torch.nn.BatchNorm2d(co),
        torch.nn.ReLU(inplace=True)
    )

def encoder_conv(ci, co):
  return torch.nn.Sequential(
        torch.nn.MaxPool2d(2),
        conv3x3_bn(ci, co),
        conv3x3_bn(co, co),
    )

class deconv(torch.nn.Module):
    def __init__(self, ci, co):
        super(deconv, self).__init__()
        self.upsample = torch.nn.ConvTranspose2d(ci, co, 2, stride=2)
        self.conv1 = conv3x3_bn(2*co, co)
        self.conv2 = conv3x3_bn(co, co)
    
    def forward(self, x1, x2):
        x1 = self.upsample(x1)
        diffX = x2.size()[2] - x1.size()[2]
        diffY = x2.size()[3] - x1.size()[3]
        x1 = F.pad(x1, (diffX, 0, diffY, 0))
        x = torch.cat([x2, x1], dim=1)
        x = self.conv1(x)
        x = self.conv2(x)
        return x
    
class SegmentationModel(nn.Module):
    def __init__(self, num_classes=6, encoder='resnet18', in_channels=3, pretrained=True):
        super().__init__()
        self.encoder = timm.create_model(encoder, pretrained=pretrained, in_chans=in_channels, features_only=True)
        decoder_channels = self.encoder.feature_info.channels()[::-1]
        self.decoder = nn.ModuleList(
            [deconv(decoder_channels[i], decoder_channels[i+1]) for i in range(len(decoder_channels)-1)]
        )
        extra_channels = decoder_channels[-1] // 2  
        self.conv_last = nn.Sequential(
            torch.nn.ConvTranspose2d(decoder_channels[-1], extra_channels, 2, stride=2),
            conv3x3_bn(extra_channels, extra_channels),
            conv3x3_bn(extra_channels, extra_channels),
            torch.nn.Conv2d(extra_channels, num_classes, 1)
        )

    def forward(self, x):
        f = self.encoder(x)
        x = f[-1]
        for i, d in enumerate(self.decoder):
            x = d(x, f[-i-2])
        x = self.conv_last(x)
        return x
    


from .AuthRepo import AuthRepo
from .APIRepo import APIRepo
from .AuthAPIRepo import AuthAPIRepo
from .DatasetsAPIRepo import DatasetsAPIRepo
from .FilesAPIRepo import FilesAPIRepo
from .ModelsAPIRepo import ModelsAPIRepo
from .STACAPIRepo import STACAPIRepo
from .FEAPIRepo import FEAPIRepo


import json

from ...repos import STACAPIRepo

def api_status():
    repo = STACAPIRepo()
    data, error = repo.status()
    if error:
        raise Exception(error)
    return data

def retrieve_stac_collections():
    repo = STACAPIRepo()
    data, error = repo.collections()
    if error:
        raise Exception(error)
    return data

def retrieve_stac_collection(collection_name):
    repo = STACAPIRepo()
    data, error = repo.collection(collection_name)
    if error:
        raise Exception(error)
    return data

def retrieve_stac_items(collection_id):
    repo = STACAPIRepo()
    data, error = repo.items(collection_id)
    if error:
        raise Exception(error)
    return data

def retrieve_stac_item(collection_id, item_id):
    repo = STACAPIRepo()
    data, error = repo.item(collection_id, item_id)
    if error:
        raise Exception(error)
    return data

def search_stac_items(collection_id, query = None):
    repo = STACAPIRepo()
    if query is None:
        data, error = repo.search_columns(collection_id)
        if error:
            raise Exception(error)
        return data
    data, error = repo.search(collection_id, str(query))
    if error:
        raise Exception(error)
    return json.loads(data)
    

def search_stac_columns(collection_id):
    repo = STACAPIRepo()
    data, error = repo.search_columns(collection_id)
    if error:
        raise Exception(error)
    return data



import lightning as L
import torch.nn as nn
import torch
from torchmetrics.image import PeakSignalNoiseRatio, StructuralSimilarityIndexMeasure
import os 

from .model import ESRT

class ESRTModule(L.LightningModule):
	def __init__(self, upscale = 2, pretrained = None):
		super().__init__()
		self.save_hyperparameters()
		self.model = ESRT(upscale=upscale)
		self.loss = nn.L1Loss()
		self.train_psnr = PeakSignalNoiseRatio(data_range=1.0)
		self.train_ssim = StructuralSimilarityIndexMeasure(data_range=1.0)
		self.val_psnr = PeakSignalNoiseRatio(data_range=1.0)
		self.val_ssim = StructuralSimilarityIndexMeasure(data_range=1.0)
		if pretrained:
			self.load_pretrained(pretrained)

	def forward(self, x):
		return self.model(x)
	
	def predict(self, x):
		self.eval()
		with torch.no_grad():
			return self.model(x.to(self.device))
	
	def training_step(self, batch, batch_idx):
		x, y = batch
		y_hat = self(x)
		loss = self.loss(y_hat, y)
		psnr = self.train_psnr(y_hat, y)
		ssim = self.train_ssim(y_hat, y)
		self.log("train_loss", loss, prog_bar=True)
		self.log("train_psnr", psnr, prog_bar=True)
		self.log("train_ssim", ssim, prog_bar=True)
		return loss
	
	def validation_step(self, batch, batch_idx):
		x, y = batch
		y_hat = self(x)
		loss = self.loss(y_hat, y)
		psnr = self.val_psnr(y_hat, y)
		ssim = self.val_ssim(y_hat, y)
		self.log("val_loss", loss, prog_bar=True, sync_dist=True)
		self.log("val_psnr", psnr, prog_bar=True, sync_dist=True)
		self.log("val_ssim", ssim, prog_bar=True, sync_dist=True)

	def configure_optimizers(self):
		optimizer = torch.optim.AdamW(self.model.parameters(), lr=3e-4, weight_decay=1e-4)
		scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=200, gamma=0.5)
		return [optimizer], [scheduler]
	
	def load_pretrained(self, path):
		if os.path.isfile(path):
			print("===> loading model '{}'".format(path))
			checkpoint = torch.load(path)['state_dict']
			model_dict = self.model.state_dict()
			pretrained_dict = {k.replace('model.', ''): v for k, v in checkpoint.items()}
			for k, v in model_dict.items():
				if k not in pretrained_dict:
					print(k)
			self.model.load_state_dict(pretrained_dict, strict=True)
		else:
			print("===> no models found at '{}'".format(path))



from ..repos import FEAPIRepo


def retrieve_pipelines(name=None, limit=None):
    api_repo = FEAPIRepo()
    data, error = api_repo.retrieve_pipelines(name, limit)
    if data and not error:
        models = [d["name"] for d in data] if data else []
        return models
    return []


def retrieve_pipeline(name):
    repo = FEAPIRepo()
    data, error = repo.retrieve_pipeline(name)
    if error:
        raise Exception(error)
    return data


import torch 
import skimage.io as io
from einops import rearrange
import numpy as np
from skimage.transform import resize
from skimage import exposure

class Dataset(torch.utils.data.Dataset):
	def __init__(
			self, 
			hr_paths, 
			lr_paths, 
			hr_bands, 	
			lr_bands,
			upscale = 2, 
			trans = None,
			resize = True
		):
		self.hr_paths = hr_paths
		self.lr_paths = lr_paths
		self.upscale = upscale
		self.trans = trans
		self.hr_bands = hr_bands
		self.lr_bands = lr_bands
		self.resize = resize

	def __len__(self):
		return len(self.hr_paths)
	
	def __getitem__(self, idx):
		try:
			hr = io.imread(self.hr_paths[idx])[..., self.hr_bands].astype(np.float32)
		except Exception as e:
			print(f"Error loading HR image {self.hr_paths[idx]}: {e}")
			return self.__getitem__(0)
		try:
			lr = io.imread(self.lr_paths[idx])[..., self.lr_bands].astype(np.float32)
		except Exception as e:
			print(f"Error loading LR image {self.lr_paths[idx]}: {e}")
			return self.__getitem__(0)
		if self.trans: # flips and crops
			trans = self.trans(image=hr, image2=lr)
			hr = trans["image"]
			lr = trans["image2"]
		hr = exposure.match_histograms(hr, lr, channel_axis=-1)
		# normalize
		hr = np.clip(hr / 4000, 0, 1)
		lr = np.clip(lr / 4000, 0, 1)
		# center crop satellogic at 380x380 (original size is 384x384) and lr at 38x38
		hr = hr[2:-2, 2:-2, :]
		h, w = lr.shape[:2]
		lr = lr[h//2-19:h//2+19, w//2-19:w//2+19, :]
		# resize according to the upscale factor
		target_size = 38 * self.upscale
		# make sure 380 is divisible by the upscale factor
		# if 380 % self.upscale != 0:
		# 	raise ValueError(f"Target size {target_size} is not divisible by the upscale factor {self.upscale}")
		if self.resize:
			hr = resize(hr, (target_size, target_size), order=3, anti_aliasing=False)
		# channels first
		hr = rearrange(hr, "h w c -> c h w")
		lr = rearrange(lr, "h w c -> c h w")
		return lr, hr#, hr0


"""
Search imagery
"""

from typing import Union, List
from datetime import datetime

from .sentinelhub import (
    SHClient,
    get_default_parameters,
    evaluate_sentinel_parameters,
    SHParameters,
    supports_cloud_coverage,
)


def search_sentinel_imagery(
    time_interval: Union[str, datetime, List[Union[str, datetime]]],
    bounding_box: List[Union[int, float]],
    collection_id: str,
) -> None:
    """
    Search Sentinel imagery
    """
    evaluate_sentinel_parameters(
        time_interval, bounding_box, collection_id, output_needed=False
    )
    parameters = get_default_parameters(collection_id)
    client = SHClient(parameters.BASE_URL)
    return client.search_data(bounding_box, time_interval, parameters)


def advanced_imagery_search(
    time_interval: Union[str, datetime, List[Union[str, datetime]]],
    bounding_box: List[Union[int, float]],
    parameters: SHParameters,
) -> None:
    """
    Advanced imagery search
    """
    evaluate_sentinel_parameters(
        time_interval, bounding_box, output_needed=False, parameters=parameters
    )

    if (
        supports_cloud_coverage(parameters.DATA_COLLECTION.api_id)
        and parameters.MAX_CLOUD_COVERAGE
    ):
        parameters.FILTER = "eo:cloud_cover < " + str(parameters.MAX_CLOUD_COVERAGE)

    client = SHClient(parameters.BASE_URL)
    return client.search_data(bounding_box, time_interval, parameters)



from pathlib import Path

from eotdl.files.metadata import Metadata


def test_save_metadata_supports_unicode_authors(tmp_path):
    metadata = Metadata(
        authors=["Janez Čuk", "María Šimek"],
        license="CC-BY-4.0",
        source="https://example.com",
        name="unicode-dataset",
        description="Unicode-safe description",
    )

    readme_path = Path(metadata.save_metadata(tmp_path))
    content = readme_path.read_text(encoding="utf-8")

    assert "Janez Čuk" in content
    assert "María Šimek" in content



"""
Statistics
"""

from datetime import datetime
from typing import Union, List
import pandas as pd

from .sentinelhub import (
    SHClient,
    evaluate_sentinel_parameters,
    SHParameters,
    stats_to_df,
)


def get_statistics(
    time_interval: Union[str, datetime, List[Union[str, datetime]]],
    bounding_box: List[Union[int, float]],
    parameters: SHParameters,
    evalscript: str,
) -> None:
    """
    Download Sentinel imagery
    """
    client = SHClient(parameters.BASE_URL)

    parameters.EVALSCRIPT = evalscript
    evaluate_sentinel_parameters(
        time_interval, bounding_box, parameters=parameters, output_needed=False
    )
    req = client.requestStatistics(time_interval, bounding_box, parameters)
    stat_data = client.download_statistical_data(req)
    dfs = [stats_to_df(stats) for stats in stat_data]
    return pd.concat(dfs)



import lightning as L
import segmentation_models_pytorch as smp
import torch
import timm

class UnsupervisedModule(L.LightningModule):

    def __init__(self, config):
        super().__init__()
        self.save_hyperparameters(config)
        self.backbone = timm.create_model(
            self.hparams.backbone,
            pretrained=False,
            features_only=True,
            in_chans=self.hparams.in_chans
        )
        self.head = torch.nn.Sequential(
            torch.nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            torch.nn.Flatten(),
            torch.nn.Linear(
                self.backbone.feature_info.channels(-1), self.hparams.mlp_dim),
            torch.nn.BatchNorm1d(self.hparams.mlp_dim),
            torch.nn.ReLU(),
            torch.nn.Linear(self.hparams.mlp_dim, self.hparams.mlp_dim),
            torch.nn.BatchNorm1d(self.hparams.mlp_dim),
            torch.nn.ReLU(),
            torch.nn.Linear(self.hparams.mlp_dim, self.hparams.mlp_dim)
        )
        self.l = 5e-3

    def forward(self, x):
        return self.head(self.backbone(x)[-1])

    def training_step(self, batch, batch_idx):
        # two randomly augmented versions of x
        x1, x2 = batch
        # compute representations
        z1 = self(x1)
        z2 = self(x2)
        # normalize repr. along the batch dimension
        N, D = z1.shape
        z1_norm = (z1 - z1.mean(0)) / z1.std(0)  # NxD
        z2_norm = (z2 - z2.mean(0)) / z2.std(0)  # NxD
        # cross-correlation matrix
        c = (z1_norm.T @ z2_norm) / N  # DxD
        # loss
        c_diff = (c - torch.eye(D, device=self.device)).pow(2)  # DxD
        # multiply off-diagonal elems of c_diff by lambda
        d = torch.eye(D, dtype=bool)
        c_diff[~d] *= self.l
        loss = c_diff.sum()
        self.log('loss', loss, prog_bar=True)
        return loss

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters())
        max_epochs = self.trainer.max_epochs
        scheduler = torch.optim.lr_scheduler.MultiStepLR(
            optimizer, milestones=[int(0.6*max_epochs), int(0.8*max_epochs)])
        return [optimizer], [scheduler]

class EuroSATModule(L.LightningModule):

    def __init__(self, config, backbone):
        super().__init__()
        self.save_hyperparameters(config)
        self.backbone = backbone
        if self.hparams.freeze:
            for param in self.backbone.parameters():
                param.requires_grad = False
        self.head = torch.nn.Sequential(
            torch.nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            torch.nn.Flatten(),
            torch.nn.Linear(
                self.backbone.feature_info[-1]['num_chs'], self.hparams.num_classes)
        )
        self.criterion = torch.nn.CrossEntropyLoss()

    def forward(self, x):
        if self.hparams.freeze:
            self.backbone.eval()
            with torch.no_grad():
                features = self.backbone(x)
        else:
            features = self.backbone(x)
        return self.head(features[-1])

    def training_step(self, batch, batch_idx):
        loss, acc = self.shared_step(batch)
        self.log('loss', loss)
        self.log('acc', acc, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        loss, acc = self.shared_step(batch)
        self.log('val_loss', loss, prog_bar=True)
        self.log('val_acc', acc, prog_bar=True)

    def shared_step(self, batch):
        x, y = batch
        logits = self(x)
        loss = self.criterion(logits, y)
        acc = (torch.argmax(logits, axis=1) == y).sum().item() / y.size(0)
        return loss, acc

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters())



import torch.nn as nn
import timm
import torch

class ClassificationModel(nn.Module):
    def __init__(self, num_classes=6, encoder='resnet18', in_channels=3, pretrained=True, ssl=None):
        super().__init__()
        self.model = timm.create_model(encoder, pretrained=pretrained, in_chans=in_channels, num_classes=num_classes)
        if ssl:
            print(f'Loading SSL checkpoint from {ssl}')
            # Load SSL checkpoint
            checkpoint = torch.load(ssl, map_location='cpu')
            state_dict = checkpoint['state_dict']
            # Remove module prefix if present (from DataParallel/DistributedDataParallel)
            new_state_dict = {}
            for k, v in state_dict.items():
                if k.startswith('module.'):
                    k = k[7:]  # remove 'module.' prefix
                new_state_dict[k] = v
            # Load the state dict
            self.model.load_state_dict(new_state_dict, strict=False)

    def forward(self, x):
        return self.model(x)




import os
import pytest

from eotdl.curation.stac.extensions import add_ml_extension, MLDatasetQualityMetrics
from eotdl.curation.stac.extensions.ml_dataset import MLDatasetExtension

from .fixtures import tmp_stac_catalog, tmp_stac_catalog_labels


def test_add_ml_extension(tmp_stac_catalog):
    add_ml_extension(tmp_stac_catalog)

    assert MLDatasetExtension.has_extension(tmp_stac_catalog) is True


def test_add_ml_extension_export(tmp_stac_catalog):
    add_ml_extension(tmp_stac_catalog, destination="tmp/exported_catalog")

    assert os.path.exists("tmp/exported_catalog/catalog.json")


def test_add_ml_extension_parameters(tmp_stac_catalog):
    add_ml_extension(
        tmp_stac_catalog,
        name="Q2 Dataset",
        tasks=["segmentation"],
        inputs_type=["satellite imagery"],
        annotations_type="raster",
        version="0.1.0",
    )

    assert tmp_stac_catalog.extra_fields["ml-dataset:name"] == "Q2 Dataset"
    assert tmp_stac_catalog.extra_fields["ml-dataset:tasks"] == ["segmentation"]
    assert tmp_stac_catalog.extra_fields["ml-dataset:inputs-type"] == [
        "satellite imagery"
    ]
    assert tmp_stac_catalog.extra_fields["ml-dataset:annotations-type"] == "raster"
    assert tmp_stac_catalog.extra_fields["ml-dataset:version"] == "0.1.0"


def test_add_ml_extension_splits_without_labels(tmp_stac_catalog):
    with pytest.raises(AttributeError):
        add_ml_extension(
            tmp_stac_catalog,
            splits=True,
            splits_collection_id="labels",
            name="Q2 Dataset",
            tasks=["segmentation"],
            inputs_type=["satellite imagery"],
            annotations_type="raster",
            version="0.1.0",
        )


def test_add_ml_extension_splits(tmp_stac_catalog_labels):
    add_ml_extension(
        tmp_stac_catalog_labels,
        splits=True,
        splits_collection_id="labels",
        name="Q2 Dataset",
        tasks=["segmentation"],
        inputs_type=["satellite imagery"],
        annotations_type="raster",
        version="0.1.0",
    )

    assert tmp_stac_catalog_labels.extra_fields["ml-dataset:splits"] == (
        "Training",
        "Validation",
        "Test",
    )

    collection = tmp_stac_catalog_labels.get_child("labels")
    assert "ml-dataset:split-items" in collection.extra_fields


def test_add_ml_extension_quality_metrics(tmp_stac_catalog_labels):
    add_ml_extension(
        tmp_stac_catalog_labels,
        splits=True,
        splits_collection_id="labels",
        name="Q2 Dataset",
        tasks=["segmentation"],
        inputs_type=["satellite imagery"],
        annotations_type="raster",
        version="0.1.0",
    )

    MLDatasetQualityMetrics.calculate(tmp_stac_catalog_labels)

    assert "ml-dataset:quality-metrics" in tmp_stac_catalog_labels.extra_fields
    quality_metrics = tmp_stac_catalog_labels.extra_fields["ml-dataset:quality-metrics"]
    assert isinstance(quality_metrics, list) is True
    assert len(quality_metrics) == 2
    assert quality_metrics[0]["name"] == "spatial-duplicates"
    assert quality_metrics[1]["name"] == "classes-balance"



import lightning as L
import torch
import torch.nn as nn
from torchmetrics.classification import MultilabelAccuracy

from .model import ClassificationModel

class MultiLabelClassificationModule(L.LightningModule):
    def __init__(self, num_classes, encoder='resnet18', in_channels=3, pretrained=True, lr=1e-3, ssl=None):
        super().__init__()
        self.save_hyperparameters()
        self.model = ClassificationModel(num_classes, encoder, in_channels, pretrained, ssl)
        self.loss = nn.BCEWithLogitsLoss()
        self.train_acc = MultilabelAccuracy(num_labels=num_classes)
        self.val_acc = MultilabelAccuracy(num_labels=num_classes)

    def forward(self, x):
        return self.model(x)
    
    def predict(self, x):
        self.eval()
        with torch.no_grad():
            preds = self.model(x.to(self.device))
            return torch.sigmoid(preds, dim=1)
    
    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self(x)
        loss = self.loss(y_hat, y)
        self.train_acc(y_hat, y)
        self.log('train_loss', loss, prog_bar=True)
        self.log('train_acc', self.train_acc, prog_bar=True)
        return loss
    
    def validation_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self(x)
        loss = self.loss(y_hat, y)
        self.val_acc(y_hat, y)
        self.log('val_loss', loss, prog_bar=True)
        self.log('val_acc', self.val_acc, prog_bar=True)
        return loss
    
    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.hparams.lr)




from unittest import mock

import pandas as pd

import eotdl.datasets.stage as stage_module


def test_stage_dataset_downloads_all_assets_in_parallel_path(monkeypatch, tmp_path):
    dataset = {
        "id": "dataset-id",
        "name": "my-dataset",
        "metadata": {
            "authors": ["Albert"],
            "license": "CC-BY-4.0",
            "source": "https://example.com",
            "description": "desc",
            "thumbnail": "",
        },
        "versions": [{"version_id": 1}],
    }
    gdf = pd.DataFrame(
        [
            {"assets": {"asset": {"href": "https://example.com/asset-1"}}},
            {
                "assets": {
                    "rgb": {"href": "https://example.com/asset-2"},
                    "nir": {"href": "https://example.com/asset-3"},
                }
            },
        ]
    )
    staged_urls = []

    class FakeFilesAPIRepo:
        def stage_file(
            self,
            dataset_or_model_id,
            file_name,
            user,
            path,
            endpoint="datasets",
            progress=False,
        ):
            catalog_path = tmp_path / "catalog.v1.parquet"
            catalog_path.write_bytes(b"catalog")
            return str(catalog_path)

        def stage_file_url(self, url, path, user):
            staged_urls.append(url)
            return str(tmp_path / "downloaded-file")

    monkeypatch.setenv("EOTDL_STAGE_WORKERS", "4")

    with (
        mock.patch("eotdl.auth.auth.auth", return_value={"id_token": "token"}),
        mock.patch.object(stage_module, "retrieve_dataset", return_value=dataset),
        mock.patch.object(stage_module, "FilesAPIRepo", return_value=FakeFilesAPIRepo()),
        mock.patch.object(stage_module.gpd, "read_parquet", return_value=gdf),
    ):
        stage_module.stage_dataset("my-dataset", assets=True, path=str(tmp_path / "downloads"))

    assert len(staged_urls) == 3
    assert set(staged_urls) == {
        "https://example.com/asset-1",
        "https://example.com/asset-2",
        "https://example.com/asset-3",
    }




from .temporal_utils import compute_temporal_extent
from .spatial_utils import  buffer_geometry

import os
from typing import List
import geopandas as gpd

import pandas as pd
from typing import List, Dict, Any

try:
    import openeo_gfmap
except ImportError:
    print("openeo_gfmap is not installed, please install it with 'pip install openeo_gfmap'")

from openeo_gfmap.manager.job_splitters import split_job_s2grid, append_h3_index

def create_utm_patch(
    row: Any,
    parent_crs: str,
    start_date: str,
    nb_months: int,
    pixel_size: int,
    resolution: int
) -> Dict[str, Any]:
    """
    Process a single row from a GeoDataFrame to extract metadata and buffer geometry.
    
    Parameters:
        row: A single row of a GeoDataFrame.
        start_date: Start date for the temporal extent (ISO format string).
        nb_months: Number of months for the temporal extent.
        buffer_distance: Buffer distance in meters for the geometry.
    
    Returns:
        A dictionary with metadata and processed geometry.
    """
    # Extract the polygon geometry from the row
    # Create a single-row GeoDataFrame
    gdf = gpd.GeoDataFrame(
        [row], 
        crs=parent_crs, 
        geometry="geometry"
    )
    
    buffer = int(pixel_size*resolution/2)
    # Apply the buffer_geometry function
    buffered_gdf = buffer_geometry(gdf, buffer=buffer, resolution = resolution)

    # Compute the temporal extent
    temporal_extent = compute_temporal_extent(start_date, nb_months)

    # Return the processed data
    return {
        "fid": row.get("fid"),  # Include any relevant identifier
        "geometry": buffered_gdf.iloc[0].geometry,
        "crs": buffered_gdf.crs.to_string(),
        "temporal_extent": temporal_extent,
    }

def process_patch_geodataframe(
    geodataframe: gpd.GeoDataFrame,
    start_date: str,
    nb_months: int,
    buffer_distance: int,
    resolution: int
) -> gpd.GeoDataFrame:
    """
    Process a GeoDataFrame to generate buffered geometries and metadata for each row.
    
    Parameters:
        geodataframe: The input GeoDataFrame with geometries to process.
        start_date: Start date for the temporal extent (ISO format string).
        nb_months: Number of months for the temporal extent.
        buffer_distance: Buffer distance in meters for the geometry.
        resolution: Spatial resolution for geometry rounding.
    
    Returns:
        A processed GeoDataFrame with buffered geometries and additional metadata.
    """
    results = []  # List to store processed rows
    
    for _, row in geodataframe.iterrows():
        # Process each row and collect the result
        result = create_utm_patch(row, geodataframe.crs, start_date, nb_months, buffer_distance, resolution)
        results.append(result)
    
    # Convert the list of results into a GeoDataFrame
    processed_gdf = gpd.GeoDataFrame(
        results, 
        geometry="geometry", 
        crs=result['crs']
    )
    
    return processed_gdf

def process_geodataframe(
    geodataframe: gpd.GeoDataFrame,
    start_date: str,
    nb_months: int,
    extra_cols: List[str] = [],

) -> gpd.GeoDataFrame:
    """
    Process a GeoDataFrame to generate buffered geometries and metadata for each row.
    
    Parameters:
        geodataframe: The input GeoDataFrame with geometries to process.
        start_date: Start date for the temporal extent (ISO format string).
        nb_months: Number of months for the temporal extent.
        buffer_distance: Buffer distance in meters for the geometry.
        resolution: Spatial resolution for geometry rounding.
    
    Returns:
        A processed GeoDataFrame with buffered geometries and additional metadata.
    """
    results = []  # List to store processed rows
    geodataframe = geodataframe.to_crs(epsg=4326)
    
    for _, row in geodataframe.iterrows():
    
        # Compute the temporal extent
        temporal_extent = compute_temporal_extent(start_date, nb_months)

        # Return the processed data
        result =  {
            # "fid": row.get("fid"),  # Include any relevant identifier
            "geometry": row.geometry,
            "crs": geodataframe.crs.to_string(),
            "temporal_extent": temporal_extent,
            "s1_weekly_statistics_url": row.get("s1_weekly_statistics_url"),
            "s2_weekly_statistics_url": row.get("s2_weekly_statistics_url"),
            **{col: row[col] for col in extra_cols}
        }  
        
        results.append(result)
    
    # Convert the list of results into a GeoDataFrame
    processed_gdf = gpd.GeoDataFrame(
        results, 
        geometry="geometry", 
        crs=result['crs']
    )
    
    return processed_gdf

def split_geodataframe_by_s2_grid(base_gdf: gpd.GeoDataFrame, max_points:int, grid_resolution:int = 3) -> List[gpd.GeoDataFrame]:
    """Append H3 index and split into smaller job dataframes."""
    original_crs = base_gdf.crs
    # Append H3 index, which will change the CRS temporarily
    base_gdf = append_h3_index(base_gdf, grid_resolution=grid_resolution)
    # Transform back to the original CRS
    h3_gdf = base_gdf.to_crs(original_crs)
    split_gdf = split_job_s2grid(h3_gdf, max_points=max_points)
    return split_gdf

#TODO evaluate need
def generate_featurecollection_dataframe(split_jobs: List[gpd.GeoDataFrame]) -> pd.DataFrame:
    """Create a DataFrame from split jobs with essential information for each job, including feature count."""
    job_data = []
    for job in split_jobs:

        # Ensure the temporal_extent field exists and handle missing data
        temporal_extent = job.temporal_extent.iloc[0] if 'temporal_extent' in job.columns and job.temporal_extent.iloc[0] else None
        
        # Handle missing S2 and H3 information gracefully
        s2_tile = job.tile.iloc[0] if 'tile' in job.columns and job.tile.iloc[0] else None
        h3index = job.h3index.iloc[0] if 'h3index' in job.columns and job.h3index.iloc[0] else None
        
        # Extract CRS as string
        crs = job.crs.to_string() if job.crs else None
        
        # Count the number of features (rows) in the GeoDataFrame
        feature_count = len(job)
        
        # Serialize the entire GeoDataFrame to GeoJSON (including geometry and attributes)
        job_json = job.to_json()

        # Append all information, including feature count
        job_data.append({
            'temporal_extent': temporal_extent,
            'geometry': job_json,  # The entire GeoDataFrame serialized to GeoJSON
            's2_tile': s2_tile,
            'h3index': h3index,
            'crs': crs,
            'feature_count': feature_count  # Include the feature count for the job
        })

    # Return the DataFrame with all job metadata
    return pd.DataFrame(job_data)


def process_and_create_advanced_patch_jobs(
    gdf: gpd.GeoDataFrame, 
    start_date: str, 
    nb_months: int, 
    pixel_size: int, 
    resolution: int, 
    max_points: int
) -> pd.DataFrame:
    """
    A wrapper function to process geospatial data and generate a job metadata DataFrame.
    
    Args:
        gdf (GeoDataFrame): The input GeoDataFrame with geometries.
        start_date (str): The start date for the temporal extent (ISO format).
        nb_months (int): Number of months for the temporal extent.
        buffer_distance (int): Buffer distance in meters for geometry.
        resolution (int): Spatial resolution for geometry rounding.
        max_points (int): Maximum number of points per job for splitting the grid.
    
    Returns:
        pd.DataFrame: A DataFrame containing the job metadata.
    """
    # Step 1: Process GeoDataFrame in which we create patches and temporal info
    processed_gdf = process_patch_geodataframe(
        gdf, start_date, nb_months, pixel_size, resolution
    )

    # Step 2: Split processed GeoDataFrame into smaller jobs by Sentinel-2 grid
    split_jobs = split_geodataframe_by_s2_grid(processed_gdf, max_points)

    # Step 3: Generate the job metadata DataFrame
    job_df = generate_featurecollection_dataframe(split_jobs)

    return job_df








from datetime import datetime, timezone
import io
import os
from pathlib import Path
import tempfile
import time

from eotdl.datasets.retrieve import retrieve_dataset_files
import numpy as np
import pytest

from eotdl.datasets import ingest_dataset, retrieve_dataset
import rasterio


os.environ["EOTDL_API_URL"] = "http://localhost:8000/"


def write_readme(path: Path, name: str):
    readme_text = f"""---
name: {name}
authors: 
  - Derek van de Ven
license: free
source: https://github.com/earthpulse/eotdl/blob/main/tutorials/notebooks/02_ingesting.ipynb
---

# {name}

This file is nonsensical data used for load testing. It should not be stored on the web. 
"""

    with open(path / name / "README.md", "w") as outfile:
        outfile.write(readme_text)


def generate_fake_tif_dataset(
    path: Path, n_tifs: int = 10, tif_size: int = 256, name: str = "test_set"
):
    files_to_upload = []
    for count in range(n_tifs):
        files_to_upload.append(f"{name}/FakeFolder/Fake_tif{count}.tif")

    for rel_path in files_to_upload:
        file_path = path / rel_path
        file_path.parent.mkdir(parents=True, exist_ok=True)

        width, height = tif_size, tif_size
        data = np.full((height, width, 3), 256, dtype=np.uint32)

        with rasterio.open(
            file_path,
            "w",
            driver="GTiff",
            height=height,
            width=width,
            count=3,
            dtype="uint32",
            crs="EPSG:4326",
            transform=rasterio.transform.from_origin(0, height, 1, 1),
        ) as dst:
            dst.write(data[:, :, 0], 1)
            dst.write(data[:, :, 1], 2)
            dst.write(data[:, :, 2], 3)

    write_readme(path, name)


def generate_fake_dataset(
    path: Path, size_mb: int = 10, n_files: int = 5, name: str = "test_set"
):
    files_to_upload = []
    for count in range(n_files):
        files_to_upload.append(
            f"{name}/FakeFolder/Fake_tif{count}.tif",
        )

    total_size_bytes = size_mb * 1024 * 1024
    file_size = total_size_bytes // len(files_to_upload)

    for rel_path in files_to_upload:
        file_path = path / rel_path
        file_path.parent.mkdir(parents=True, exist_ok=True)

        # Write random binary data to simulate `.tif` files
        with open(file_path, "wb") as f:
            f.write(os.urandom(file_size))  # Random binary data

    write_readme(path, name)


@pytest.mark.parametrize(
    "total_size, n_files",
    [
        (1, 10),
        # (1e1, 10),
        # (1e3, 1), # 1 files amounting to 1 GB,
        # (1e3, 20), # 20 files amounting to 1 GB
        # (1e4, 100), # 100 10 MB files
        # (3e4, 100), # 10 times 1 GB
    ],
)
def test_load(setup_mongo, setup_minio, total_size, n_files):
    name = f"LoadTest-{int(total_size)}MB"
    with tempfile.TemporaryDirectory(prefix="loadtest_") as tmpdir:
        tmpdir = Path(tmpdir)
        generate_fake_dataset(
            path=tmpdir, size_mb=int(total_size), n_files=n_files, name=name
        )

        # upload
        start_time = time.time()
        ingest_dataset(path=str(tmpdir / name))

        upload_duration = time.time() - start_time
        print(f"Upload for {name} took {upload_duration:.2f} seconds.")

        with open("eotdl/tests/load/upload_times.log", "a") as log_file:
            log_file.write(
                f"{total_size} MB made of {n_files} files took {upload_duration:.2f} seconds.\n"
            )

        # assert dataset ingested
        dataset = retrieve_dataset(name=name)
        assert dataset["name"] == name
        assert round(dataset["versions"][0]["size"] / (1024 * 1024), 2) == total_size

        # check files in minio
        minio_files = []
        minio_sizes = []
        for obj in setup_minio.list_objects("eotdl-test", recursive=True):
            minio_files.append(obj.object_name)
            minio_sizes.append(obj.size)
            assert obj.object_name.endswith(("tif", "md", "parquet"))

        assert len(minio_files) == n_files + 2
        assert round(sum(minio_sizes) / (1024 * 1024)) == total_size


@pytest.mark.parametrize(
    "tif_size, n_tifs",
    [
        # (256, 100),
        # (512, 100),
        # (1024, 100),
        (2048, 200),
    ],
)
def test_load_tifs(setup_mongo, setup_minio, tif_size, n_tifs):
    name = f"LoadTest-{int(tif_size)}"
    with tempfile.TemporaryDirectory(prefix="loadtest_") as tmpdir:
        tmpdir = Path(tmpdir)
        generate_fake_tif_dataset(
            path=tmpdir, tif_size=tif_size, n_tifs=n_tifs, name=name
        )

        # upload
        start_time = time.time()
        ingest_dataset(path=str(tmpdir / name))

        upload_duration = time.time() - start_time
        print(f"Upload for {name} took {upload_duration:.2f} seconds.")

        # assert dataset ingested
        dataset = retrieve_dataset(name=name)
        assert dataset["name"] == name

        with open("eotdl/tests/load/upload_times_tifs.log", "a") as log_file:
            log_file.write(
                f"{tif_size}x {tif_size} made of {n_tifs} files took {upload_duration:.2f} seconds. ({round(dataset['versions'][0]['size'] / (1024 * 1024), 2)} mb)\n"
            )

        # check files in minio
        minio_files = []
        minio_sizes = []
        for obj in setup_minio.list_objects("eotdl-test", recursive=True):
            minio_files.append(obj.object_name)
            minio_sizes.append(obj.size)
            assert obj.object_name.endswith(("tif", "md", "parquet"))

            if obj.object_name.endswith(".tif"):
                file = setup_minio.get_object("eotdl-test", obj.object_name)
                file_content = file.read()
                with io.BytesIO(file_content) as byte_file:
                    with rasterio.open(byte_file) as src:
                        assert src.crs == "EPSG:4326"
                        assert src.width == src.height == tif_size
                        assert src.dtypes == ("uint32", "uint32", "uint32")

                        for band in range(1, 4):
                            colour = src.read(band)
                            assert set(colour.flatten()) == {np.uint32(256)}

        assert len(minio_files) == n_tifs + 2



from ..auth import with_auth
from ..repos import FilesAPIRepo
from ..datasets.retrieve import retrieve_dataset
from ..models.retrieve import retrieve_model
from ..fe.retrieve import retrieve_pipeline

@with_auth
def get_file_content_url(filename, dataset_or_model_name, endpoint, user):
    if endpoint == "datasets":
        dataset_or_model_id = retrieve_dataset(dataset_or_model_name)['id']
    elif endpoint == "models":
        dataset_or_model_id = retrieve_model(dataset_or_model_name)['id']
    elif endpoint == "pipelines":
        dataset_or_model_id = retrieve_pipeline(dataset_or_model_name)['id']
    else:
        raise Exception("Invalid endpoint (datasets, models or pipelines)")
    repo = FilesAPIRepo()
    return repo.generate_file_content_url(filename, dataset_or_model_id, user, endpoint)


import requests

from ..repos import APIRepo

class DatasetsAPIRepo(APIRepo):
    def __init__(self, url=None):
        super().__init__(url)

    def retrieve_datasets(self, name, limit):
        url = self.url + "datasets"
        if name is not None:
            url += "?match=" + name
        if limit is not None:
            if name is None:
                url += "?limit=" + str(limit)
            else:
                url += "&limit=" + str(limit)
        response = requests.get(url)
        return self.format_response(response)
    
    def retrieve_private_datasets(self, user):
        url = self.url + "datasets/private"
        response = requests.get(url, headers=self.generate_headers(user))
        return self.format_response(response)
    
    def retrieve_dataset(self, name):
        response = requests.get(self.url + "datasets?name=" + name)
        return self.format_response(response)
    
    def retrieve_private_dataset(self, name, user):
        response = requests.get(self.url + "datasets/private?name=" + name, headers=self.generate_headers(user))
        return self.format_response(response)
    
    def get_dataset_by_id(self, dataset_id):
        response = requests.get(self.url + "datasets/" + dataset_id)
        return self.format_response(response)
    
    def create_dataset(self, metadata, user, private=False):
        if private:
            metadata["visibility"] = "private"
        response = requests.post(
            self.url + "datasets",
            json=metadata,
            headers=self.generate_headers(user),
        )
        return self.format_response(response)

    def complete_ingestion(self, dataset_id, version, size, user):
        response = requests.post(
            self.url + "datasets/complete/" + dataset_id,
            json={"version": version, "size": size},
            headers=self.generate_headers(user),
        )
        return self.format_response(response)
    
    def deactivate_dataset(self, dataset_id, user):
        response = requests.patch(
            self.url + "datasets/" + dataset_id + "/deactivate",
            headers=self.generate_headers(user),
        )
        return self.format_response(response)



from ..repos import DatasetsAPIRepo
from ..auth import with_auth
from .retrieve import retrieve_dataset

def update_dataset(dataset_id, metadata, content, user):
    repo = DatasetsAPIRepo()
    data, error = repo.update_dataset(
        dataset_id,
        metadata.authors,
        metadata.source,
        metadata.license,
        metadata.thumbnail,
        content,
        user,
    )
    if error:
        raise Exception(error)
    return data

@with_auth
def deactivate_dataset(dataset_name, user):
    dataset = retrieve_dataset(dataset_name)
    repo = DatasetsAPIRepo()
    data, error = repo.deactivate_dataset(dataset['id'], user)
    if error:
        raise Exception(error)
    return data


from datetime import datetime, timedelta
import multiprocessing
from tqdm import tqdm
import os
import numpy as np

from ..access import search_sentinel_imagery
from ..tools import bbox_from_centroid
from .utils import _find_matches

def find_sentinel_matches_by_centroid(centroid, date, time_buffer, width, height, collection_id="sentinel-2-l2a"):
    dates = [(date - timedelta(days=time_buffer/2)).strftime('%Y-%m-%d'),
             (date + timedelta(days=time_buffer/2)).strftime('%Y-%m-%d')]
    custom_bbox = bbox_from_centroid(x=centroid.y, y=centroid.x, pixel_size=10, width=width, height=height)
    sentinel_matches = list(search_sentinel_imagery(dates, custom_bbox, collection_id))
    return sentinel_matches

def find_sentinel_matches_by_bb(bb, date, time_buffer, collection_id="sentinel-2-l2a"):
    if isinstance(date, str):
        date = datetime.strptime(date, '%Y-%m-%d')
    dates = [(date - timedelta(days=time_buffer/2)).strftime('%Y-%m-%d'),
             (date + timedelta(days=time_buffer/2)).strftime('%Y-%m-%d')]
    sentinel_matches = list(search_sentinel_imagery(dates, bb, collection_id))
    return sentinel_matches

def find_matches(
    gdf, 
    time_buffer, 
    width, 
    height, 
    collection_id,
    num_cores=multiprocessing.cpu_count()
):
    from concurrent.futures import ProcessPoolExecutor
    with ProcessPoolExecutor(max_workers=num_cores) as pool:
        args = [(ix, row.geometry.centroid, row.date, time_buffer, width, height, collection_id) for ix, row in gdf.iterrows()]
        with tqdm(total=len(args)) as progress:
            futures = []
            for arg in args:
                future = pool.submit(_find_matches, arg) # enviamos la tupla de argumentos
                future.add_done_callback(lambda p: progress.update())
                futures.append(future)
            results = []
            for future in futures:
                result = future.result()
                results.append(result)
    return results

def find_matches_chunked(
    path, 
    input_table_name,
    output_table_name,
    chunk_size, 
    num_chunks, # set to -1 to process the entire table
    seed = 2025, 
    shuffle = True,
    collections = {
        "sentinel-2-l2a": {
            'column': 's2_matches',
            'time_buffer': 30,
            'width': 384,
            'height': 384
        },
        "sentinel-1-grd": {
            'column': 's1_matches',
            'time_buffer': 6,
            'width': 384,
            'height': 384
        }
    },
    num_cores=multiprocessing.cpu_count(),
    remove_chunks=True
):
    import pyarrow.parquet as pq
    import geopandas as gpd
    from shapely import wkb
    import pandas as pd
    import gc
    import shutil

    # shutil.rmtree(path + 'chunks', ignore_errors=True)
    os.makedirs(path + 'chunks', exist_ok=True)

    print("Reading items... ", end="", flush=True)
    table = pq.read_table(path + input_table_name)
    if shuffle:
        num_rows = table.num_rows
        np.random.seed(seed)  # Set the seed for repeatability
        shuffled_indices = np.random.permutation(num_rows)
        table = table.take(shuffled_indices)
    print("Done", flush=True)

    if num_chunks < 0:
        num_chunks = (len(table) + chunk_size - 1) // chunk_size

    for i in range(num_chunks):
        if os.path.exists(path + f'chunks/chunk-{chunk_size}-{seed}-{i+1}.parquet'):
            print(f"Chunk {i+1}/{num_chunks} already exists, skipping...")
            continue
        
        start_idx = i * chunk_size
        end_idx = min(start_idx + chunk_size, len(table))

        print(f"Processing chunk {i+1}/{num_chunks}... ")
        chunk = table.slice(start_idx, end_idx - start_idx)
        df = chunk.to_pandas()
        df['geometry'] = df['geometry'].apply(wkb.loads)
        gdf_sampled = gpd.GeoDataFrame(df, geometry='geometry')

        for collection_id, param in collections.items():
            print(f"Finding {collection_id} matches... ", end="", flush=True)
            results = find_matches(gdf_sampled, param['time_buffer'], param['width'], param['height'], collection_id, num_cores)
            matches_df = pd.DataFrame(results, columns=['index', param['column']]).set_index('index')
            gdf_sampled[param['column']] = matches_df[param['column']]
            print("Done", flush=True)

        print("Saving results... ", end="", flush=True)
        gdf_sampled.to_parquet(path + f'chunks/chunk-{chunk_size}-{seed}-{i+1}.parquet')
        print("Done", flush=True)

        del chunk, df, gdf_sampled, results, matches_df
        gc.collect()


    print("Merging chunks... ", end="", flush=True)
    # chunk_files = sorted(glob(f"{path}/chunks/chunk-{chunk_size}-{seed}-*.parquet"))
    chunk_files = [path + f'chunks/chunk-{chunk_size}-{seed}-{i+1}.parquet' for i in range(num_chunks)]
    writer = None
    try:
        for chunk_file in chunk_files:
            if not os.path.exists(chunk_file):
                continue
            table = pq.read_table(chunk_file)
            if writer is None:
                writer = pq.ParquetWriter(path + output_table_name, table.schema)
            writer.write_table(table)
    finally:
        if writer is not None:
            writer.close()
    if remove_chunks:
        shutil.rmtree(path + 'chunks', ignore_errors=True)
        shutil.rmtree(path + '.chunks', ignore_errors=True)
    print("Done", flush=True)

def download_matches():
    pass